package Classifier::MailParse;

# ---------------------------------------------------------------------------------------------
#
# MailParse.pm --- Parse a mail message or messages into words
#
# ---------------------------------------------------------------------------------------------

use strict;
use warnings;
use locale;
use Classifier::WordMangle;

#----------------------------------------------------------------------------
# new
#
# Class new() function
#----------------------------------------------------------------------------

sub new 
{
    my $type = shift;
    my $self;

    # Used to mangle words into the right shape for classification

    $self->{mangle} = new Classifier::WordMangle;
  
    # Hash of word frequences

    $self->{words}  = {};

    # Total word cout

    $self->{msg_total} = 0;
    
    # Debug messages?

    $self->{debug}     = 0;
    
    # Internal use for keeping track of a line without touching it
    
    $self->{ut}        = '';
    
    # Specifies the parse mode, 1 means color the output
    
    $self->{color}     = 0;
    
    # This will store the from, to and subject from the last parse
    
    $self->{from}      = '';
    $self->{to}        = '';
    $self->{subject}   = '';

    return bless $self, $type;
}

# ---------------------------------------------------------------------------------------------
#
# un_base64
#
# Decode a line of base64 encoded data
#
# $line     A line of base64 encoded data
#
# ---------------------------------------------------------------------------------------------

sub un_base64 
{
    my ($self, $line) = @_;
    my $result;
    
    $line =~ s/=+$//; 
    $line =~ s/[\r\n]//g; 
    $line =~ tr|A-Za-z0-9+/| -_|;

    $result = join'', map( unpack("u", chr(32 + length($_)*3/4) . $_), $line =~ /(.{1,196})/gs);
    $result =~ s/\x00//g;
    
    return $result;
}

# ---------------------------------------------------------------------------------------------
#
# update_word
#
# Updates the word frequency for a word 
#
# ---------------------------------------------------------------------------------------------

sub update_word 
{
    my ($self, $word, $encoded, $before, $after) = @_;

    print "--- $word ($before) ($after)\n" if ($self->{debug});
    
    my $mword = $self->{mangle}->mangle($word);
    
    if ( $mword ne '' )  {
        if ( $self->{color} ) {
            my $color = $self->{bayes}->get_color($mword);
            if ( $encoded == 0 )  {
                $after = '&' if ( $after eq '>' );
                $self->{ut} =~ s/($before)\Q$word\E($after)/$1<b><font color=$color>$word<\/font><\/b>$2/;
            } else {
                $self->{ut} .= "Found in encoded data <font color=$color>$word<\/font>\r\n";
            }
        } else {
            $self->{words}{$mword} += 1;
            $self->{msg_total}     += 1;

            print "--- $word ($self->{words}{$mword}) ($before) ($after)\n" if ($self->{debug});
        }
    }
}

# ---------------------------------------------------------------------------------------------
#
# add_line
#
# Parses a single line of text and updates the word frequencies
#
# $bigline      The line to split into words and add to the word counts
# $encoded      1 if the line was found in encoded text (base64)
#
# ---------------------------------------------------------------------------------------------

sub add_line 
{
    my ($self, $bigline, $encoded) = @_;
    my $p = 0;
    
    # If the line is really long then split at every 1k and feed it to the parser below
    
    while ( $p < length($bigline) ) {
        my $line = substr($bigline, $p, 1024);
        
        # Pull out any email addresses in the line that are marked with <> and have an @ in them

        while ( $line =~ s/<([[:alpha:]0-9\-_\.]+?@([[:alpha:]0-9\-_\.]+?))>// )  {
            update_word($self, $1, $encoded, ';', '&');
            update_word($self, $2, $encoded, '@', '&');
        }

        # Grab domain names

        while ( $line =~ s/(([[:alpha:]][[:alpha:]0-9\-_]+\.){2,})([[:alpha:]0-9\-_]+)([^[:alpha:]0-9\-_]|$)/$4/ )  {
            update_word($self, "$1$3", $encoded, '', '');
        }

        # Grab IP addresses

        while ( $line =~ s/(([12]?\d{1,2}\.){3}[12]?\d{1,2})// )  {
            update_word($self, "$1", $encoded, '', '');
        }

        # Only care about words between 3 and 45 characters since short words like
        # an, or, if are too common and the longest word in English (according to
        # the OED) is pneumonoultramicroscopicsilicovolcanoconiosis

        while ( $line =~ s/([[:alpha:]][[:alpha:]\']{0,44})[_\-,\.\"\'\)\?!:;\/&]{0,5}([ \t\n\r]|$)/ / ) {
            update_word($self,$1, $encoded, '', '[_\-,\.\"\'\)\?!:;\/ &\t\n\r]') if (length $1 >= 3);
        }
        
        $p += 1024;
    }
}

# ---------------------------------------------------------------------------------------------
#
# update_tag
#
# Extract domain names from inside tags
#
# $tag     The tag name
# $arg     The arguments
#
# ---------------------------------------------------------------------------------------------

sub update_tag 
{
    my ($self, $tag, $arg) = @_;

    $tag =~ s/[\r\n]//g;
    $arg =~ s/[\r\n]//g;

    print "HTML tag $tag with argument " . $arg . "\n" if ($self->{debug});

    if ( ( $tag =~ /^img$/i ) || ( $tag =~ /^frame$/i ) ) {
        return update_word( $self, $3, 0, $1, $4 ) if ( $arg =~ /src[ \t]*=[ \t]*[\"\']?http:\/(\/(.+:)?)([^ \/\">]+)([ \/\">]|$)/i );
    }

    if ( $tag =~ /^a$/i )  {
        return update_word( $self, $3, 0, $1, $4 ) if ( $arg =~ /href[ \t]*=[ \t]*[\"\']?http:\/(\/(.+:)?)([^ \/\">]+)([ \/\">]|$)/i );
    }

    if ( $tag =~ /^form$/i )  {
        return update_word( $self, $3, 0, $1, $4 ) if ( $arg =~ /action[ \t]*=[ \t]*([\"\']?(http:\/\/)?)([^ \/\">]+)([ \/\">]|$)/i );
    }
}

# ---------------------------------------------------------------------------------------------
#
# parse_stream
#
# Read messages from a file stream and parse into a list of words and frequencies
#
# $file     The file to open and parse
#
# ---------------------------------------------------------------------------------------------

sub parse_stream 
{
    my ($self, $file) = @_;
    
    # This will contain the mime boundary information in a mime message

    my $mime     = '';
    
    # Contains the encoding for the current block in a mime message

    my $encoding = '';

    # Clear the word hash

    my $content_type = '';

    # Used to return a colorize page
    
    my $colorized = '';

    my $in_html_tag = 0;
    my $html_tag    = '';
    my $html_arg    = '';

    $self->{words}     = {};
    $self->{msg_total} = 0;
    $self->{from}      = '';
    $self->{to}        = '';
    $self->{subject}   = '';
    $self->{ut}        = '';

    $colorized .= "<tt>" if ( $self->{color} );

    open MSG, "<$file";
    binmode MSG;
    
    # Read each line and find each "word" which we define as a sequence of alpha
    # characters
    
    while (<MSG>) {
        my $read = $_;

        # For the Mac we do further splitting of the line at the CR characters
        
        while ( $read =~ s/(.*?[\r\n]+)// )  {
            my $line = $1;

            next if ( !defined($line) );
            
            print ">>> $line" if $self->{debug};

            if ( $self->{color} )  {
                my $splitline = $line;    
                $splitline =~ s/([^\r\n]{100,120} )/$1\r\n/g;
                $splitline =~ s/([^ \r\n]{120})/$1\r\n/g;

                if ( !$in_html_tag )  {
                    $colorized .= $self->{ut} if ( $self->{ut} ne '' );
                    
                    $self->{ut} = '';
                }

                $splitline =~ s/</&lt;/g;
                $splitline =~ s/>/&gt;/g;
                $splitline =~ s/\t/&nbsp;&nbsp;&nbsp;&nbsp;/g;
                $self->{ut} .= $splitline;
            }

            # If we are in a mime document then spot the boundaries

            if ( ( $mime ne '' ) && ( $line =~ /$mime/ ) ) {
                print "Hit mime boundary\n" if $self->{debug};
                $encoding = '';
                next;
            }

            # If we are doing base64 decoding then look for suitable lines and remove them 
            # for decoding

            if ( $encoding =~ /base64/i ) {
                my $decoded = '';
                $self->{ut} = '' if $self->{color};
                print "ba> [$line]" if $self->{debug};
                while ( ( $line =~ /^([A-Za-z0-9+\/]{4}){1,48}[\n\r]*/ ) || ( $line =~ /^[A-Za-z0-9+\/]+=+?[\n\r]*/ ) ) {
                    print "64> $line" if $self->{debug};
                    $decoded    .= un_base64( $self, $line );
                    if ( $decoded =~ /[^[:alpha:]\-\.]$/ )  {
                        if ( $self->{color} ) {
                            my $splitline = $line;    
                            $splitline =~ s/([^\r\n]{120})/$1\r\n/g;
                            $self->{ut} = $splitline;
                        }
                        add_line( $self, $decoded, 1 );
                        $decoded = '';
                        if ( $self->{color} )  {
                            if ( $self->{ut} ne '' )  {
                                $colorized .= $self->{ut};
                                $self->{ut} = '';
                            }
                        }
                    }

                    last if ( !($line = <MSG>) );
                }

                add_line( $self, $decoded, 1 );
            }

            next if ( !defined($line) );
            $content_type = 'text/html' if ( $line =~ /<html>/i );

            # Transform some escape characters

            if ( $encoding =~ /quoted\-printable/ ) {
                $line =~ s/=[\r\n]*$/=\n/;
                while ( ( $line =~ /=([0-9A-F][0-9A-F])/i ) != 0 ) {
                    my $from = "=$1";
                    my $to   = chr(hex("0x$1"));
                    $line =~ s/$from/$to/g;
                }
            }

            # Remove HTML comments
            
            $line =~ s/<!--.*?-->//g;

            # Remove HTML tags completely

            if ( $content_type =~ /html/ )  {
                if ( $in_html_tag )  {
                    if ( $line =~ s/(.*?)>/ / ) {
                        $html_arg .= $1;
                        $in_html_tag = 0;
                        $html_tag =~ s/=\n ?//g if ( $encoding =~ /quoted\-printable/ );
                        $html_arg =~ s/=\n ?//g if ( $encoding =~ /quoted\-printable/ );
                        update_tag( $self, $html_tag, $html_arg );
                        $html_tag = '';
                        $html_arg = '';
                    } else {
                        $html_arg .= " " . $line;
                        $line = '';
                        next;
                    }
                }
                    
                while ( $line =~ s/<[\/]?([A-Za-z]+)([^>]*?)>/ / )  {
                    update_tag( $self, $1, $2 );
                }
                
                if ( $line =~ s/<([^ >]+)([^>]*)$// )  {
                    $html_tag = $1;
                    $html_arg = $2;
                    $in_html_tag = 1;
                }
            }

            # If we have an email header then just keep the part after the :

            if ( $line =~ /^([A-Za-z-]+): ?([^\n\r]*)/ )  {
                my $header   = $1;
                my $argument = $2;

                print "Header ($header) ($argument)\n" if ($self->{debug});

                # Handle the From, To and Cc headers and extract email addresses
                # from them and treat them as words

                if ( $header =~ /(From|To|Cc)/i ) {
                    if ( $header =~ /From/ )  {
                        $encoding     = '';
                        $content_type = '';
                        $self->{from} = $argument if ( $self->{from} eq '' ) ;
                    }

                    $self->{to} = $argument if ( ( $header =~ /To/i ) && ( $self->{to} eq '' ) );
                    
                    while ( $argument =~ s/<([[:alpha:]0-9\-_\.]+?@([[:alpha:]0-9\-_\.]+?))>// )  {
                        update_word($self, $1, 0, ';', '&');
                        update_word($self, $2, 0, '@', '&');
                    }

                    while ( $argument =~ s/([[:alpha:]0-9\-_\.]+?@([[:alpha:]0-9\-_\.]+))// )  {
                        update_word($self, $1, 0, '', '');
                        update_word($self, $2, 0, '@', '');
                    }

                    add_line( $self, $argument, 0 );
                    next;
                }

                $self->{subject} = $argument if ( ( $header =~ /Subject/ ) && ( $self->{subject} eq '' ) );

                # Look for MIME

                if ( $header =~ /Content-Type/i )  {
                    if ( $argument =~ /multipart\//i ) {
                        my $boundary = $argument;
                        
                        $boundary = <MSG> if ( !( $argument =~ /boundary=\"(.*)\"/ )); 

                        if ( $boundary =~ /boundary=\"(.*)\"/ )  {
                            print "Set mime boundary to $1\n" if $self->{debug};

                            $mime = $1;
                            $mime =~ s/(\+|\/|\?|\*|\||\(|\)|\[|\]|\{|\}|\^|\$|\.)/\\$1/g;
                        }
                    }

                    $content_type = $argument;
                    next;
                }

                # Look for the different encodings in a MIME document, when we hit base64 we will
                # do a special parse here since words might be broken across the boundaries

                if ( $header =~ /Content-Transfer-Encoding/i ) {
                    $encoding = $argument;
                    print "Setting encoding to $encoding\n" if $self->{debug};
                    next;
                }

                # Some headers to discard

                next if ( $header =~ /(Thread-Index|X-UIDL|Message-ID|X-Text-Classification|X-Mime-Key)/i );

                add_line( $self, $argument, 0 );
            } else {
                add_line( $self, $line, 0 );
            }
        }
    }

    close MSG;
    
    if ( $self->{color} )  {
        $colorized .= $self->{ut} if ( $self->{ut} ne '' );

        $colorized .= "</tt>";
        $colorized =~ s/[\r\n]+$//;
        $colorized =~ s/[\r\n]+/__BREAK__/g;
        $colorized =~ s/__BREAK__/<br>/g;
        
        return $colorized;
    }
}

1;
